﻿using NUnit.Framework;
using System;
using System.IO;
using System.Xml;

namespace Faz.SideWinderSC.Logic
{
    /// <summary>
    /// Tests the <see cref="Mapping"/> class.
    /// </summary>
    [TestFixture]
    public class MappingTest
    {
        /// <summary>
        /// Tests the <see cref="Mapping.Keys"/> property.
        /// </summary>
        [Test]
        public static void Keys()
        {
            Mapping instance = new Mapping();
            Assert.IsNotNull(instance.Keys);
            Assert.AreEqual(0, instance.Keys.Count);
            Assert.IsFalse(instance.Keys.IsReadOnly);
        }

        /// <summary>
        /// Tests the <see cref="Mapping.WriteXml"/> method.
        /// </summary>
        [Test]
        public static void WriteXml()
        {
            Mapping instance = new Mapping();

            Assert.Throws<ArgumentNullException>(() => instance.WriteXml(null));

            using (MemoryStream stream = new MemoryStream())
            using (TextReader reader = new StreamReader(stream))
            using (XmlWriter writer = XmlWriter.Create(stream, new XmlWriterSettings() { ConformanceLevel = ConformanceLevel.Fragment }))
            {
                instance.WriteXml(writer);
                writer.Flush();
                stream.Position = 0;
                Assert.AreEqual("<Mapping />", reader.ReadToEnd());

                stream.Position = 0;
                instance.Name = "test";
                instance.Button = SwscButton.Button2;
                instance.Keys.Add(ConsoleKey.F2);
                instance.Keys.Add(ConsoleKey.F1);
                instance.WriteXml(writer);
                writer.Flush();
                stream.Position = 0;
                Assert.AreEqual(
                    "<Mapping Button=\"Button2\"><Name>test</Name><Key>F2</Key><Key>F1</Key></Mapping>",
                    reader.ReadToEnd());
            }
        }

        /// <summary>
        /// Tests the <see cref="Mapping.ReadXml"/> method.
        /// </summary>
        [Test]
        public static void ReadXml_Empty()
        {
            Mapping instance = new Mapping();

            Assert.Throws<ArgumentNullException>(() => instance.ReadXml(null));

            using (MemoryStream stream = new MemoryStream())
            using (TextWriter writer = new StreamWriter(stream))
            {
                writer.Write("<?xml version=\"1.0\" encoding=\"utf-8\" ?><Mapping />");
                writer.Flush();
                stream.Position = 0;
                using (XmlReader reader = XmlReader.Create(stream, new XmlReaderSettings()))
                {
                    instance.ReadXml(reader);
                }

                Assert.AreEqual((SwscButton)0, instance.Button);
                Assert.AreEqual(0, instance.Keys.Count);
                Assert.AreEqual(null, instance.Name);
            }
        }

        /// <summary>
        /// Tests the <see cref="Mapping.ReadXml"/> method.
        /// </summary>
        [Test]
        public static void ReadXml_Complete()
        {
            Mapping instance = new Mapping();

            using (MemoryStream stream = new MemoryStream())
            using (TextWriter writer = new StreamWriter(stream))
            {
                writer.Write("<Mapping Button=\"Button2\"><Name>test mapping</Name><Key>Tab</Key><Key>F1</Key></Mapping>");
                writer.Flush();
                stream.Position = 0;
                using (XmlReader reader = XmlReader.Create(stream, new XmlReaderSettings()))
                {
                    instance.ReadXml(reader);
                }

                Assert.AreEqual(SwscButton.Button2, instance.Button);
                Assert.AreEqual("test mapping", instance.Name);
                CollectionAssert.AreEquivalent(new[] { ConsoleKey.Tab, ConsoleKey.F1 }, instance.Keys);
            }
        }

        /// <summary>
        /// Tests the <see cref="Mapping.ReadXml"/> method.
        /// </summary>
        [Test]
        public static void ReadXml_Partial1()
        {
            Mapping instance = new Mapping();

            using (MemoryStream stream = new MemoryStream())
            using (TextWriter writer = new StreamWriter(stream))
            {
                writer.Write("<Mapping Button=\"Button2\"><Key>Tab</Key></Mapping>");
                writer.Flush();
                stream.Position = 0;
                using (XmlReader reader = XmlReader.Create(stream, new XmlReaderSettings()))
                {
                    instance.ReadXml(reader);
                }

                Assert.AreEqual(SwscButton.Button2, instance.Button);
                Assert.AreEqual(null, instance.Name);
                CollectionAssert.AreEquivalent(new[] { ConsoleKey.Tab }, instance.Keys);
            }
        }

        /// <summary>
        /// Tests the <see cref="Mapping.ReadXml"/> method.
        /// </summary>
        [Test]
        public static void ReadXml_Partial2()
        {
            Mapping instance = new Mapping();

            using (MemoryStream stream = new MemoryStream())
            using (TextWriter writer = new StreamWriter(stream))
            {
                writer.Write("<Mapping Button=\"Button2\"><Name /></Mapping>");
                writer.Flush();
                stream.Position = 0;
                using (XmlReader reader = XmlReader.Create(stream, new XmlReaderSettings()))
                {
                    instance.ReadXml(reader);
                }

                Assert.AreEqual(SwscButton.Button2, instance.Button);
                Assert.IsNullOrEmpty(instance.Name);
                Assert.AreEqual(0, instance.Keys.Count);
            }
        }
    }
}
