﻿using System;
using System.IO;
using System.Linq;
using NUnit.Framework;
using System.Collections.Generic;

namespace Faz.SideWinderSC.Logic
{
    /// <summary>
    /// Tests the <see cref="ModelService"/> class.
    /// </summary>
    [TestFixture]
    public class ModelServiceTest
    {
        /// <summary>
        /// Tests the <see cref="ModelService.RetrieveModels"/> method.
        /// </summary>
        [Test]
        public static void RetrieveModels()
        {
            ModelService instance = new ModelService();
            var actuals = instance.RetrieveModels();

            Assert.IsNotNull(actuals);
            Assert.AreEqual(2, actuals.Count);
            Assert.IsNotNull(actuals.ElementAt(0));
            Assert.IsNotNull(actuals.ElementAt(1));
            Assert.AreEqual("Strategic", actuals.ElementAt(0).Name);
            Assert.AreEqual("Dummy", actuals.ElementAt(1).Name);
        }

        /// <summary>
        /// Tests the <see cref="ModelService.RetrieveProfiles"/> method.
        /// </summary>
        [Test]
        public static void RetrieveProfiles()
        {
            Assert.Throws<ArgumentNullException>(() => ModelService.RetrieveProfiles((Model)null));

            var actuals = ModelService.RetrieveProfiles(new Model() { Name = string.Empty });
            Assert.IsNotNull(actuals);

            Assert.AreEqual(3, actuals.Count);
            Assert.IsNotNull(actuals.ElementAt(0));
            Assert.IsNotNull(actuals.ElementAt(1));
            Assert.IsNotNull(actuals.ElementAt(2));
            Assert.AreEqual("Unit Test 1", actuals.ElementAt(0).Name);
            Assert.AreEqual("Unit Test 2", actuals.ElementAt(1).Name);
            Assert.AreEqual("Unit Test 3", actuals.ElementAt(2).Name);
        }

        /// <summary>
        /// Tests the <see cref="ModelService.RetrieveMappings"/> method.
        /// </summary>
        [Test]
        public static void RetrieveMappings_Empty()
        {
            Assert.Throws<ArgumentNullException>(() => ModelService.RetrieveMappings(null));

            Model model = new Model() { Name = string.Empty };
            Profile profile = new Profile(model) { Name = "Unit Test 1" };
            var actuals = ModelService.RetrieveMappings(profile);

            Assert.IsNotNull(actuals);
            Assert.AreEqual(0, actuals.Count);
        }

        /// <summary>
        /// Tests the <see cref="ModelService.RetrieveMappings"/> method.
        /// </summary>
        [Test]
        public static void RetrieveMappings_Complete()
        {
            Assert.Throws<ArgumentNullException>(() => ModelService.RetrieveMappings(null));

            Model model = new Model() { Name = string.Empty };
            model.DefaultMappings.Add(new Mapping() { Button = SwscButton.Button2, Name = "default 2" });
            model.DefaultMappings.Add(new Mapping() { Button = SwscButton.Button5, Name = "default 5" });
            Profile profile = new Profile(model) { Name = "Unit Test 2" };

            var actuals = ModelService.RetrieveMappings(profile);

            Assert.IsNotNull(actuals);
            Assert.AreEqual(4, actuals.Count);
            Assert.IsNotNull(actuals.ElementAt(0));
            Assert.IsNotNull(actuals.ElementAt(1));
            Assert.IsNotNull(actuals.ElementAt(2));
            Assert.IsNotNull(actuals.ElementAt(3));
            Assert.AreEqual("mapping 1", actuals.ElementAt(0).Name);
            Assert.AreEqual("mapping 2", actuals.ElementAt(1).Name);
            Assert.AreEqual("mapping 3", actuals.ElementAt(2).Name);
            Assert.AreEqual("default 5", actuals.ElementAt(3).Name);
        }

        /// <summary>
        /// Tests the <see cref="ModelService.SaveProfile"/> method.
        /// </summary>
        [Test]
        public static void SaveProfile()
        {
            ModelService instance = new ModelService();

            // test errors
            Assert.Throws<ArgumentNullException>(() => instance.SaveProfile(null));

            // test normal profile
            Profile actual = new Profile(new Model() { Name = "" });
            actual.Name = "Unit Test 2 Copy";
            actual.Mappings.Add(new MockMapping(new[] { ConsoleKey.D0 }) { Name = "mapping 1", Button = SwscButton.Button1 });
            actual.Mappings.Add(new MockMapping(new[] { ConsoleKey.A }) { Name = "mapping 2", Button = SwscButton.Button2 });
            actual.Mappings.Add(new MockMapping(new[] { ConsoleKey.Tab }) { Name = "mapping 3", Button = SwscButton.Button3 });

            if (File.Exists("Unit Test 2 Copy.profile"))
            {
                Assert.Inconclusive("The file Unit Test 2 Copy.profile already exists");
            }

            // test a first save - set up FilePath
            instance.SaveProfile(actual);
            StringAssert.EndsWith("Unit Test 2 Copy.profile", actual.FilePath);
            Assert.IsTrue(File.Exists(actual.FilePath));

            // test a second save - FilePath unmodified
            actual.Name = "Unit Test 2";
            instance.SaveProfile(actual);
            StringAssert.EndsWith("Unit Test 2 Copy.profile", actual.FilePath);
            XmlAssert.AreEqual("Unit Test 2.profile", actual.FilePath);
        }

        /// <summary>
        /// Cleans the test environment after all tests.
        /// </summary>
        [TestFixtureTearDown]
        public static void TestFixtureTearDown()
        {
            File.Delete("Unit Test 2 Copy.profile");
        }

        private class MockMapping : Mapping
        {
            public MockMapping(ICollection<ConsoleKey> keys)
            {
                foreach (ConsoleKey key in keys)
                {
                    this.Keys.Add(key);
                }
            }
        }
    }
}
