﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NUnit.Framework;
using System.IO;
using System.ComponentModel;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace Faz.SideWinderSC.Logic.Win32Api
{
    /// <summary>
    /// Tests the <see cref="HidMethods"/> class.
    /// </summary>
    [TestFixture]
    public class HidMethodsTest
    {
        /// <summary>
        /// Tests the <see cref="HidMethods.GetHidGuid"/> method.
        /// </summary>
        [Test]
        public static void GetHidGuid()
        {
            Guid actual;
            HidMethods.GetHidGuid(out actual);

            Assert.AreEqual(Guid.Parse("4d1e55b2-f16f-11cf-88cb-001111000030"), actual);
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.GetPreparsedData(SafeFileHandle, out SafePreparsedDataHandle)"/> method.
        /// </summary>
        [Test]
        public static void GetPreparsedData_Standard()
        {
            SafePreparsedDataHandle actual;

            // Tests with invalid parameters
            Assert.Throws(typeof(ArgumentNullException), () => HidMethods.GetPreparsedData(null, out actual));
            using (var file = new SafeFileHandle(IntPtr.Zero, true))
            {
                Assert.IsFalse(HidMethods.GetPreparsedData(file, out actual));
            }

            // Try to open a stream linked to the device
            string devicePath = GetDevicePath();
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            {
                Assert.IsFalse(file.IsInvalid);

                try
                {
                    Win32Assert.Test(HidMethods.GetPreparsedData(file, out actual));
                    Assert.IsFalse(actual.IsInvalid);
                }
                finally
                {
                    actual.Dispose();
                }
            }
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.GetPreparsedData(SafeFileHandle)"/> method.
        /// </summary>
        [Test]
        public static void GetPreparsedData_Wrapper()
        {
            // Tests with invalid parameters
            Assert.Throws(typeof(ArgumentNullException), () => HidMethods.GetPreparsedData(null));
            using (var file = new SafeFileHandle(IntPtr.Zero, true))
            {
                Assert.Throws(typeof(Win32Exception), () => HidMethods.GetPreparsedData(file));
            }

            // Try to open a stream linked to the device
            string devicePath = GetDevicePath();
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            {
                Assert.IsFalse(file.IsInvalid);

                using (SafePreparsedDataHandle actual = HidMethods.GetPreparsedData(file))
                {
                    Assert.IsFalse(actual.IsInvalid);
                }
            }
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.GetCaps(SafePreparsedDataHandle, out HidCaps)"/> method.
        /// </summary>
        [Test]
        public static void GetCaps_Standard()
        {
            // Test on a valid device
            string devicePath = GetDevicePath();
            IntPtr pointer;
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            using (SafePreparsedDataHandle preparsedData = HidMethods.GetPreparsedData(file))
            {
                // Save the pointer address for the next test
                pointer = preparsedData.DangerousGetHandle();

                HidCaps actual;
                HidpStatus result = HidMethods.GetCaps(preparsedData, out actual);
                Assert.AreEqual(HidpStatus.Success, result);
                Assert.IsNotNull(actual);
                Assert.AreNotEqual(0, actual.InputReportByteLength);
                Assert.AreNotEqual(0, actual.FeatureReportByteLength);
            }

            // Test with invalid data
            using (var data = new SafePreparsedDataHandle(pointer + 1, true))
            {
                HidCaps actual;
                HidpStatus result = HidMethods.GetCaps(data, out actual);
                Assert.AreEqual(HidpStatus.InvalidPreparsedData, result);
            }
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.GetCaps(SafePreparsedDataHandle)"/> method.
        /// </summary>
        [Test]
        public static void GetCaps_Wapper()
        {
            // Test on a valid device
            string devicePath = GetDevicePath();
            IntPtr pointer;
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            using (SafePreparsedDataHandle preparsedData = HidMethods.GetPreparsedData(file))
            {
                // Save the pointer address for the next test
                pointer = preparsedData.DangerousGetHandle();

                HidCaps actual = HidMethods.GetCaps(preparsedData);
                Assert.IsNotNull(actual);
                Assert.AreNotEqual(0, actual.InputReportByteLength);
                Assert.AreNotEqual(0, actual.FeatureReportByteLength);
            }

            // Test with invalid data
            using (var data = new SafePreparsedDataHandle(pointer + 1, true))
            {
                Assert.Throws(typeof(Win32Exception), () => HidMethods.GetCaps(data));
            }
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.GetFeature"/> method.
        /// </summary>
        [Test]
        public static void GetFeature()
        {
            string devicePath = GetDevicePath();
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            using (SafePreparsedDataHandle preparsedData = HidMethods.GetPreparsedData(file))
            {
                // Extract the capabilities that provides the report id and its size
                HidCaps capabilities = HidMethods.GetCaps(preparsedData);
                int count = capabilities.FeatureReportByteLength;
                byte[] buffer = new byte[count];

                // Initialize the report id
                buffer[0] = (byte)capabilities.UsagePage;

                Win32Assert.Test(HidMethods.GetFeature(file, buffer, count));
                Assert.AreEqual(capabilities.UsagePage, buffer[0]);
            }
        }

        /// <summary>
        /// Tests the <see cref="HidMethods.SetFeature"/> method.
        /// </summary>
        [Test]
        public static void SetFeature()
        {
            string devicePath = GetDevicePath();
            using (var file = Kernel32Methods.CreateFile(devicePath, Win32FileAccess.GenericRead, 0, IntPtr.Zero, FileMode.Open, Win32FileAttributes.Overlapped, IntPtr.Zero))
            using (SafePreparsedDataHandle preparsedData = HidMethods.GetPreparsedData(file))
            {
                // Extract the capabilities that provides the report id and its size
                HidCaps capabilities = HidMethods.GetCaps(preparsedData);
                int count = capabilities.FeatureReportByteLength;
                byte[] buffer = new byte[count];

                // Initialize the report id
                buffer[0] = (byte)capabilities.UsagePage;

                Win32Assert.Test(HidMethods.SetFeature(file, buffer, count));
                Assert.AreEqual(capabilities.UsagePage, buffer[0]);
            }
        }

        /// <summary>
        /// Get the path of a Swsc controller to open a valid controller
        /// This code is based on the SetupApiMethods class that is tested in an other
        /// unit test. Any error during the process raises an inconclusive status.
        /// </summary>
        /// <returns>The path of a Swsc controller</returns>
        private static string GetDevicePath()
        {
            Guid deviceClass = Guid.Parse("4d1e55b2-f16f-11cf-88cb-001111000030");
            string search = "vid_045e&pid_0033";
            string devicePath = null;
            try
            {
                using (var deviceSet = SetupApiMethods.GetClassDevs(ref deviceClass, null, IntPtr.Zero, DiGetClassFlags.DeviceInterface | DiGetClassFlags.Present))
                {
                    uint index = 0;
                    DeviceInterfaceData interfaceData = new DeviceInterfaceData();
                    DeviceInterfaceDetailData actual = new DeviceInterfaceDetailData();
                    actual.Size = 5;
                    bool result = SetupApiMethods.EnumDeviceInterfaces(deviceSet, 0, ref deviceClass, index, interfaceData);
                    while (result)
                    {
                        uint requiredSize = 0;
                        SetupApiMethods.GetDeviceInterfaceDetail(deviceSet, interfaceData, IntPtr.Zero, 0, ref requiredSize, IntPtr.Zero);
                        SetupApiMethods.GetDeviceInterfaceDetail(deviceSet, interfaceData, ref actual, requiredSize, ref requiredSize, IntPtr.Zero);

                        if (actual.DevicePath.Contains(search))
                        {
                            devicePath = actual.DevicePath;
                            break;
                        }

                        index++;
                        result = SetupApiMethods.EnumDeviceInterfaces(deviceSet, 0, ref deviceClass, index, interfaceData);
                    }
                }
            }
            catch (Exception e)
            {
                Assert.Inconclusive("Issue with SetupApisMethods calls, check their tests first - {0}", e.Message);
            }

            if (devicePath == null)
            {
                Assert.Inconclusive("No swsc controller found to execute the test");
            }

            return devicePath;
        }
    }
}
